//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_LIBRARY_CLASS_HH
#define BMP_LIBRARY_CLASS_HH

#include <set>

#include <glibmm.h>
#include <glibmm/i18n.h>

#include <boost/optional.hpp>
#include <boost/format.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>

#ifndef BMP_PLUGIN_BUILD
#  include <taglib/tag.h>
#endif //!BMP_PLUGIN_BUILD

#include "util.hh"

#include "database.hh"

namespace Bmp
{
  namespace UiPart { class Albums; class CDDA; } 

  namespace Library
  {

#include "exception.hh"

    EXCEPTION(MetadataWriteError)
    EXCEPTION(NoMetadataError)

#ifdef HAVE_HAL
    EXCEPTION(NoHALInformationError)
#endif //HAVE_HAL

    EXCEPTION(DatabaseError)

    enum DatumTags 
    {
      TS_DATUM_ID,
      TS_DATUM_DESC,  
      TS_DATUM_TABLE
    };

    enum DatumTag 
    {
      TAG_DATUM_ID,
      TAG_DATUM_ATTR_NAME,
      TAG_DATUM_ATTR_TYPE,
      TAG_DATUM_ATTR_PRECISE,
      TAG_DATUM_ATTR_V_BOOL,
      TAG_DATUM_ATTR_V_INT,
      TAG_DATUM_ATTR_V_REAL,
      TAG_DATUM_ATTR_V_STRING,
    };

    enum Datum
    {
      DATUM_LOCATION,
      DATUM_ARTIST,
      DATUM_ALBUM,
      DATUM_TITLE,
      DATUM_TRACK,
      DATUM_TIME,
      DATUM_GENRE,
      DATUM_COMMENT,
      DATUM_RATING,
      DATUM_DATE,
      DATUM_MTIME,
      DATUM_BITRATE,
      DATUM_SAMPLERATE,
      DATUM_COUNT,
      DATUM_PLAYDATE,
      DATUM_ASIN,

#ifdef HAVE_HAL

      DATUM_HAL_VOLUME_UDI,
      DATUM_HAL_DEVICE_UDI,
      DATUM_VOLUME_RELATIVE_PATH,
      DATUM_VOLUME_RELATIVE_BASE_DIR,

#endif //HAVE_HAL

      DATUM_MB_ALBUM_ARTIST,
      DATUM_MB_ALBUM_ARTIST_ID,
      DATUM_MB_ALBUM_ARTIST_SORTNAME,
      DATUM_MB_TRACK_ID,
      DATUM_MB_ALBUM_ID,
      DATUM_MB_ARTIST_ID,
      DATUM_MB_ARTIST_SORTNAME,
      DATUM_MB_RELEASE_DATE,
      DATUM_MUSICIP_PUID,

      DATUM_NEW_ITEM,
      DATUM_HASH,

      DATUM_LAST
    };
     
    class Album
    {
      public:

        typedef boost::tuple <  oustring,
                                oustring,
                                oustring,
                                oustring,
                                oustring,
                                oustring,
                                ouint       > TR;

        oustring    artist;
        oustring    album;
        oustring    asin;
        oustring    release_id;
        oustring    artist_id;
        oustring    sortname;
        ouint       tracks;
        ouint       date;
        obool       new_item;

        std::string   key;
        TR            tpl;

        static std::string  create_key (Album const& a);
        static TR           create_tpl (Album const& a);

        Album (const Bmp::DB::Row& row);
        Album () {};
    };
    typedef std::vector< Album > AlbumV;

    bool
    operator== (Album const& a,
                Album const& b);

    bool
    operator<  (Album const& a,
                Album const& b);

    void
    album_to_attrs (Bmp::Library::Album const& album,
                    Bmp::DB::VAttributes & attrs );

    class Track
    {
      public:

#ifdef HAVE_HAL
        boost::optional <Glib::ustring>   volume_udi;
        boost::optional <Glib::ustring>   device_udi;
        boost::optional <Glib::ustring>   volume_relative_path;
#endif //HAVE_HAL

        boost::optional <Glib::ustring>   location;
        boost::optional <Glib::ustring>   artist;
        boost::optional <Glib::ustring>   album;
        boost::optional <Glib::ustring>   title;
        boost::optional <Glib::ustring>   genre;
        boost::optional <Glib::ustring>   comment;
        boost::optional <Glib::ustring>   asin;
        boost::optional <Glib::ustring>   puid;
        boost::optional <Glib::ustring>   mb_album_artist;
        boost::optional <Glib::ustring>   mb_album_artist_id;
        boost::optional <Glib::ustring>   mb_album_artist_sort_name;
        boost::optional <Glib::ustring>   mb_track_id;
        boost::optional <Glib::ustring>   mb_album_id;
        boost::optional <Glib::ustring>   mb_artist_id;
        boost::optional <Glib::ustring>   mb_artist_sort_name;
        boost::optional <Glib::ustring>   mb_release_date;
        boost::optional <Glib::ustring>   hash;

        boost::optional <uint64_t>     tracknumber;
        boost::optional <uint64_t>     duration;
        boost::optional <uint64_t>     rating;
        boost::optional <uint64_t>     date;
        boost::optional <uint64_t>     mtime;
        boost::optional <uint64_t>     bitrate;
        boost::optional <uint64_t>     samplerate;
        boost::optional <uint64_t>     count;
        boost::optional <uint64_t>     playdate;

        boost::optional <bool>    new_item;

#ifdef HAVE_HAL
        Track (const Bmp::DB::Row& row, bool use_hal = true);
#else
        Track (const Bmp::DB::Row& row);
#endif //HAVE_HAL

        Track () {};
    };
    typedef std::vector< Track > TrackV;

    class UTrack
    {
      public:

        boost::optional <Glib::ustring>   location_cur;
        boost::optional <Glib::ustring>   location_new;
        boost::optional <Glib::ustring>   artist;
        boost::optional <Glib::ustring>   album;
        boost::optional <Glib::ustring>   title;
        boost::optional <Glib::ustring>   genre;
        boost::optional <Glib::ustring>   comment;

        boost::optional <Glib::ustring>   asin;
        boost::optional <Glib::ustring>   puid;

        boost::optional <Glib::ustring>   mb_album_artist;
        boost::optional <Glib::ustring>   mb_album_artist_id;
        boost::optional <Glib::ustring>   mb_album_artist_sort_name;
        boost::optional <Glib::ustring>   mb_track_id;
        boost::optional <Glib::ustring>   mb_album_id;
        boost::optional <Glib::ustring>   mb_artist_id;
        boost::optional <Glib::ustring>   mb_artist_sort_name;
        boost::optional <Glib::ustring>   mb_release_date;

        boost::optional <Glib::ustring>   hash;

#ifdef HAVE_HAL
        boost::optional <Glib::ustring>   volume_udi;
        boost::optional <Glib::ustring>   device_udi;
        boost::optional <Glib::ustring>   volume_relative_path;
#endif //HAVE_HAL

        boost::optional <uint64_t>     tracknumber;
        boost::optional <uint64_t>     date;
        boost::optional <uint64_t>     rating;
        boost::optional <uint64_t>     mtime;
        boost::optional <uint64_t>     count;
        boost::optional <uint64_t>     playdate;

        boost::optional <bool>    new_item;

        UTrack () {}

        UTrack (Track const& track)
        {
          location_cur                = track.location;
          artist                      = track.artist;
          album                       = track.album;
          title                       = track.title;
          genre                       = track.genre;
          comment                     = track.comment;

          asin                        = track.asin;
          puid                        = track.puid;

#ifdef HAVE_HAL
          volume_udi                  = track.volume_udi;
          device_udi                  = track.device_udi;
          volume_relative_path        = track.volume_relative_path;
#endif //HAVE_HAL

          mb_album_artist             = track.mb_album_artist;
          mb_album_artist_id          = track.mb_album_artist_id;
          mb_album_artist_sort_name   = track.mb_album_artist_sort_name;
          mb_track_id                 = track.mb_track_id;
          mb_album_id                 = track.mb_album_id;
          mb_artist_id                = track.mb_artist_id;
          mb_artist_sort_name         = track.mb_artist_sort_name;

          tracknumber                 = track.tracknumber;
          date                        = track.date;
          rating                      = track.rating;
          mtime                       = track.mtime;
          playdate                    = track.playdate;
          new_item                    = track.new_item;
        }
    };

    typedef std::vector < UTrack > UTrackV;

#if 0
    struct CollectionAttribute
    {
      Datum             datum;
      bool              precise;
      DB::Variant       value;
    };
    typedef std::vector < CollectionAttribute > CollectionAttributes;
#endif

    class Library
    {
      private:

#include "exception.hh"

          EXCEPTION(NoMetadataTaglibError)
          EXCEPTION(NoMetadataGstError)

          void    
          get_metadata_row      (Glib::ustring const& uri,
                                 Bmp::DB::Row&        row) const;

      public:

          typedef sigc::signal <void, int> ModifyStart;
          typedef sigc::signal <void>      ModifyStep;
          typedef sigc::signal <void>      ModifyEnd;
          typedef sigc::signal <void>      SignalProcessing;
          typedef sigc::signal <void, Glib::ustring, Bmp::Library::UTrack > SignalRowUpdated;

          SignalProcessing& signal_processing_start ();
          SignalProcessing& signal_processing_end   ();
          SignalRowUpdated& signal_row_updated      ();

          ModifyStart& signal_modify_start ();
          ModifyStep&  signal_modify_step  ();
          ModifyEnd&   signal_modify_end   ();

          Library ();
          ~Library ();

          void query (DB::VAttributes const&  attrs,
                      DB::VRows            &  vector) const;

          void new_approve  ();
          void new_approve  (Album const& a);
          void new_drop_all (); 

          bool new_exist    (); 

          void del      ( DB::VAttributes  const&   attrs);
          int  add      ( Glib::ustring    const&   uri);

          void mod      ( UTrackV &                 tracks,
                          bool                      retag     = true,
                          bool                      db_update = true,
                          bool                      relocate  = true);

          void get      ( Glib::ustring     const&  uri,
                          Track &                   track) const;

          void project  ( std::string       const&  project,
                          DB::VRows &               vector,
                          DB::VAttributes   const&  attributes) const;
 
          bool  album_exists            (Album const& album);
          bool  album_has_new_items     (Album const& album) const;
          Album album_get               (Album const& a) const;
          void  get_albums              (DB::VRows & vector) const;

#ifdef HAVE_HAL
          bool  volume_exists           (std::string const& volume_udi, std::string const& device_udi) const;
#endif //HAVE_HAL

#if 0
          bool collection_exists (std::string const& name) const;
          void collection_create (std::string const& name, Glib::ustring const& desc, CollectionAttributes const& attrs); 
#endif
                                
          static std::string metadatum_id (Datum datum); 

          GHashTable * get_hashtable (Glib::ustring const& uri) const;
          ::Bmp::DB::DatumDefine get_datum_definition (Datum datum) const;

    protected:

          friend class Bmp::UiPart::Albums;
          friend class Bmp::UiPart::CDDA;

#ifndef BMP_PLUGIN_BUILD
          void metadata_set_common   (UTrack const& track, TagLib::Tag* tag);
#endif //!BMP_PLUGIN_BUILD

          void metadata_get_taglib   (Glib::ustring const& uri, DB::Row & row) const;
          void metadata_set_taglib   (Glib::ustring const& uri, UTrack  & track);

          void metadata_get_gst      (Glib::ustring const& uri, DB::Row &row) const;
          //void metadata_set_gst      (Glib::ustring const& uri, UTrack & track);
          
    private:

          bool m_has_new_items;
          bool m_new_items_modified;

          void recreate_db ();

          SignalProcessing signal_processing_start_;
          SignalProcessing signal_processing_end_;
          SignalRowUpdated signal_row_updated_;

          ModifyStart  signal_modify_start_;
          ModifyStep   signal_modify_step_;
          ModifyEnd    signal_modify_end_;

          DB::ValueMap  map_main;
          DB::ValueMap  map_albums;
          DB::ValueMap  map_va;
          DB::ValueMap  map_volumes;
          DB::Database *db_main;

#if 0 
          DB::ValueMap  map_tags;
          DB::ValueMap  map_tag;
          DB::Database *db_tags;

          DB::ValueMap  map_cdda;
          DB::Database *db_cdda;
#endif

    }; 

  } // namespace Library 

  std::string MetadatumId (Bmp::Library::Datum datum);

} // namespace Bmp 

#endif // !BMP_LIBRARY_CLASS_HH
